//  ComputationalAssembler.java

import swarm.Globals;
import swarm.defobj.Zone;
import swarm.objectbase.SwarmObjectImpl;

import swarm.collections.ListImpl;
import swarm.collections.ListIndex;


/**
 * The ComputationalAssembler class instances make 
 * computational processes; we have a computationa assembler instance for
 * each unit
 *
 * @author Pietro Terna 
 */
public class ComputationalAssembler extends ComputationalAssemblerBasic{

    /**
     * the constructor for ComputationalAssembler
     */

    public ComputationalAssembler (Zone aZone, OrderDistiller od, ESFrameModelSwarm mo)
    {
	// Call the constructor for the parent class.
	super(aZone, od, mo);

    }

    /** drop the computational assembler */
    public void drop(){

	InformationRepository.addToHibernatedComputationalAssemblerList(this);
    }

    /** computational operations with code -12xx (to be kept in 
     *  ComputationalAssembler and not to be transferred in 
     *  ComputationalAssemblerBasic) <br><br>
     *
     */

/**rating c1300
	*  <br><br>
	*  computational step number -1300 gets the value of the unsentProducts parameter of 	*  the unit the order is in.
	*  
	*  It checks if the unsentProducts value is >= the value in position
 	*  (4,0) of the unique received matrix (if it is, the unit needs funds) and <= 
	*  the value in position (4,1) of the same matrix (if it is, the default of the
 	*  firm is not so likely: rating evaluation).
	*  
	*  Finally it changes the  status to done;
	*/
    public void c1300(){
	float lowerUnsentBound0, upperUnsentBound0;
	int unitNumber, unitUnsent;
	Order theCurrentOrder;
	Unit theUnitTheOrderIsIn, theUnitTheOrderComesFrom;
	MemoryMatrix mU1, mU2;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1300 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	rd=4;

	lowerUnsentBound0=mm0.getValue(layer,0+rd,0+cd);
	upperUnsentBound0=mm0.getValue(layer,0+rd,1+cd);


	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();
	theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	
	/**variable theUnitTheOrderIsIn is not used now, but it is kept for future development 
	   *  of unsent orders funding
	   */
	theUnitTheOrderComesFrom=theCurrentOrder.getTheUnitTheOrderComesFrom();


	unitUnsent=theUnitTheOrderIsIn.getUnsentProducts();
	unitNumber=theUnitTheOrderIsIn.getUnitNumber();
		System.out.println("Unit " + unitNumber + " unsent orders: " + unitUnsent);

	if(unitUnsent >= lowerUnsentBound0 && unitUnsent <= upperUnsentBound0)
		System.out.println("Unit " + unitNumber + " needs funds and has a GOOD RATING");
 

	done = true;
    } // end c1300


     /** credit accounting
     *  <br><br>
     *  computational step 1321 acts via 1120 code
     *
     *  it modifies the coordinates of the cell where addenda should be applied, both in the
     *  memory matrix of the unit the order is in and of the unit the 
     *  order comes from. In this case the displacement in memory matrixes 
     *  is zero, but the feature is fully developed as a trace for future modifications. 
     *  The four coordinates (two for the memory matrix of the unit 
     *  the order is in, two for the unit the order comes from), are in 
     *  position (1,2), (1,3), (1,4), (1,5) of the second received matrix
     */
    public void c1321(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1321 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=1;

	// displacements for the unit memory matrixes coordinates
	mm1=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(1);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();
	urd0=(int) mm1.getValue(layer, 0+rd, 2);
	ucd0=(int) mm1.getValue(layer, 0+rd, 3);
	urd1=(int) mm1.getValue(layer, 0+rd, 4);
	ucd1=(int) mm1.getValue(layer, 0+rd, 5);

	checkMatrixNumber=false;
	c1120();
	checkMatrixNumber=true;
	rd=0;
	urd0=0; ucd0=0; urd1=0; ucd1=0;

    } // end c1321

  /** computational code 1311 
     *  <br><br>
     *
     *  this computational code creates a copy of the unit the order is in, in the same 
     *  position of the copied unit (the same x, y coordinates), but in another model stratum,       *  with the probability set in position (0,0) of the unique received matrix; the created
     *  unit is set in the stratum indicated in position (0,1) of the unique received matrix
     *  and is identified by the code of the original unit + an addendum indicated in position
     *  (0,2) of the unique received matrix;
     *  the unit is created, with the above probability, if the content of the unit memory
     *  matrix in position 0,0 is > the value in position (0,3) of the unique received matrix;
     *  if the destination position 
     *  in the new stratum is occupied the new unit is not created, without any advice; the 
     *  new unit is not created if the original one is in the destination stratum; the new unit 
     *  has the same visibility of the old one; 
     *  finally, the computational step changes the status to done; 
     *
     *  the memory matrix of the copied unit is shared among the original unit itself and its        *  copies, avoiding the creation of new matrixes for the copies    
     *
     */
    public Unit c1311(){
	double prob, count;
	Unit theUnitTheOrderIsIn, aUnit;
	Order theCurrentOrder;
	MemoryMatrix mU;
	int toStratum;
	float limit;

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=1 && checkMatrixNumber)
	    {
		System.out.println("Code -1311 requires one matrix; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	mm0=(MemoryMatrix) pendingComputationalSpecificationSet.
	    getMemoryMatrixAddress(0);
	layer=pendingComputationalSpecificationSet.
	    getOrderLayer();

	rd=0;

	addendum = (int) mm0.getValue(layer,0+rd,2+cd);
	prob    = mm0.getValue(layer,0+rd,0+cd);
	limit = mm0.getValue(layer,0+rd,3+cd);

	theCurrentOrder=pendingComputationalSpecificationSet.getOrder();

	// if whichUnit is not null, the unit we are in is set from outside
	if(whichUnit==null)theUnitTheOrderIsIn=theCurrentOrder.getTheUnitTheOrderIsIn();
	else theUnitTheOrderIsIn=whichUnit;

	mU=theUnitTheOrderIsIn.getMemoryMatrix();

	aUnit=null;
	if(mU.getValue(layer,0+urd,0+ucd)>limit 
	   && prob >= Globals.env.uniformDblRand.getDoubleWithMin$withMax(0, 1))
	    {
		toStratum = (int) mm0.getValue(layer,0+rd,1+cd);
		// unit not in the destination stratum
		if(toStratum != ((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
		   getStratum())
		    {

			aUnit=((ESFrameModelSwarm) theUnitTheOrderIsIn.getUnitModel()).
			    copyUnitToAnotherStratum(theUnitTheOrderIsIn,(int) 0, addendum, 
						     toStratum); // aUnit is used below
			                                         // if aUnit==null no space
                                                                 // has been found and the unit
                                                                 // has not been created

		    }
	    }

	addendum=0;

	done=true;
	return aUnit;
    } // end c1311

    
   /** computational code 1396 calls code c1199, using rd=2 to redirect
     *  the called step to the correct row of the input matrixes 
     */
    public void c1396(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1396 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=2;
	c1199();
	rd=0;

    } // end c1396

  /** computational code 1397 calls code c1199, using rd=3 to redirect
     *  the called step to the correct row of the input matrixes 
     */
    public void c1397(){

	if(pendingComputationalSpecificationSet.
	   getNumberOfMemoryMatrixesToBeUsed()!=2)
	    {
		System.out.println("Code -1397 requires 2 matrixes; " + 
				   pendingComputationalSpecificationSet.
				   getNumberOfMemoryMatrixesToBeUsed() + 
				   " found in order # " +
				   pendingComputationalSpecificationSet.
				   getOrderNumber());
		MyExit.exit(1);
	    }

	rd=3;
	c1199();
	rd=0;

    } // end c1397
 
}
